/************************************************************************
 * This program is Copyright (C) 1986 by Jonathan Payne.  JOVE is	*
 * provided to you without charge, and with no warranty.  You may give	*
 * away copies of JOVE, including sources, provided that this notice is *
 * included in all the files.						*
 ************************************************************************/

#ifndef _TEMP_H_
#define _TEMP_H_

#ifndef TUNED
#   include "tune.h"
#endif

#ifdef Extern
RCS_H(term, "$Id: temp.h,v 14.30 1993/01/26 18:43:05 tom Exp tom $")
#endif

#ifndef Extern
#   define Extern	extern
#endif

/* The tmp file is indexed in chunks of CH_SIZE characters.  CH_SIZE is
   (1 << CH_BITS).  New lines are added to the end of the tmp file.  The
   file is not garbage collected because that would be too painful.  As a
   result, commands like Yank and Kill are really easy; basically all we
   do is make copies of the disk addresses of the lines (as opposed to
   the contents).  So, putline(buf) writes BUF to the disk and returns a
   new disk address.  Getline(addr, buf) is the opposite of putline().
   f_getputl(line, fp) reads from open FP directly into the tmp file (into
   the buffer cache (see below)) and stores the address in LINE.  This is
   used during read_file to minimize compying.

   Lines do NOT cross block bounderies in the tmp file so that accessing
   the contents of lines can be much faster.  Pointers to offsets into
   disk buffers are returned instead of copying the contents into local
   arrays and then using them.	This cuts down on the amount of copying a
   great deal, at the expense of less efficiency.  The lower bit of disk
   addresses is used for marking lines as needing redisplay done.

   There is a buffer cache of NBUF buffers (64 on !SMALL machines and the
   3 on small ones).  The blocks are stored in LRU order and each block
   is also stored in a hash table by block #.  When a block is requested
   it can quickly be looked up in the hash table.  If it's not there the
   LRU block is assigned the new block #.  If it finds that the LRU block
   is dirty (i.e., has pending IO) it syncs the WHOLE tmp file, i.e.,
   does all the pending writes.  This works much better on floppy disk
   systems, like the IBM PC, if the blocks are sorted before sync'ing. */

#ifdef SMALL
#   define CH_BITS		4
#   define MAX_BLOCKS		((16384 / BLKSIZ) << (CH_BITS + 1))
#else
#   define CH_BITS		0
#   define MAX_BLOCKS		0	/* meaning, unlimited */
#endif /* SMALL */

#if BLKSIZ == 512
#   define BLKSIZ_SHIFT		9
#endif
#if BLKSIZ == 1024
#   define BLKSIZ_SHIFT		10
#endif
#if BLKSIZ == 2048
#   define BLKSIZ_SHIFT		11
#endif
#if BLKSIZ == 4096
#   define BLKSIZ_SHIFT		12
#endif
#if BLKSIZ == 8192
#   define BLKSIZ_SHIFT		13
#endif

#if (BLKSIZ_SHIFT < 9) || (13 < BLKSIZ_SHIFT)	/* consistency check */
	#error "value of BLKSIZ must be a power of 2, in range [512 .. 8192]"
#endif

/* CH_SIZE is how big each chunk is.  For each 1 the DFree pointer
   is incremented we extend the tmp file by CH_SIZE characters.
   CH_PBLOCK is the # of chunks per block.  RND_MASK is used to mask
   off the lower order bits of the daddr to round down to the beginning
   of a block.	OFF_MASK masks off the higher order bits so we can get
   at the offset into the disk buffer.

   NOTE:  It's pretty important that these numbers be multiples of
	  2.  Be careful if you change things. */

#define CH_SIZE 		(1 << CH_BITS)
#define CH_PBLOCK		(BLKSIZ / CH_SIZE)
#define RND_MASK		(CH_PBLOCK - 1)
#define OFF_MASK		(BLKSIZ - 1)
/*- #define BNO_MASK		(MAX_BLOCKS - 1) -*/
/* - this is obsolete now that SMALL `disk_line' is unsigned. */
#define BNO_SHIFT		(BLKSIZ_SHIFT - CH_BITS)
#define blk_round(daddr)	(daddr & ~RND_MASK)
#define forward_block(daddr)	(daddr + CH_PBLOCK)
#define daddr2bno(daddr)	((daddr >> BNO_SHIFT)/* & BNO_MASK */)
#define daddr2off(daddr)	((daddr << CH_BITS) & OFF_MASK)

typedef struct block {
	short		bl_dirty;	/* block needs saving? */
	unsigned short	bl_bno;		/* block sequence number */
	struct block	*bl_next,	/* successor in LRU list */
			*bl_prev,	/* predecessor in LRU list */
			*bl_hash;	/* successor in hash list */
	char		bl_buf[BLKSIZ];	/* the buffer proper */
} Block;

extern Block	*cache_block __((Block *_(b)));

#endif /* _TEMP_H_ */

/*======================================================================
 * $Log: temp.h,v $
 * Revision 14.30  1993/01/26  18:43:05  tom
 * cleanup whitespace.
 *
 * Revision 14.26  1992/08/26  23:56:48  tom
 * add RCS directives.
 *
 */
