#+++
#  FILE
#
#	Makefile	build termcap library
#
#  KEY WORDS
#
#	libraries
#	test functions
#
#  SYNOPSIS
#
#	make		compile the library sources
#	make tests	compile sources for tests
#	make tset	compile tset
#
#  DESCRIPTION
#
#	Standard make file for building the termcap library and tests.
#
#  AUTHOR
#
#	Fred Fish
#
#  MODIFIED
#
#	9-Mar-90 Tom Hageman, AZG, Groningen
#
#  COMPILE-TIME OPTIONS
#
#	FASTCAP=1	- fast-lookup implementation (default)
#	FASTCAP=0 or
#	SLOWCAP		- conventional implementation	
#
#	HASHED		- build hashtable for faster lookup
#			  (conventional implementation; obsoleted by FASTCAP)
#
#	Operating system options -- are set automagically when possible
#
#	GEMDOS		- gemdos (Atari ST) operating system
#	MSDOS		- MS/PC Dos operating system
#
#	HARDWIRED = 1	- compiles in a hard-coded termcap entry
#			  currently for MSDOS and GEMDOS only.
#
#	HARDWIRED = -1	- compiles in hardwired entry, alongside the
#			  normal tgetent.
#---

#CFLAGS	=	-O -I.	# for jove/mytermcap

DEBUG =		-y
LDEBUG =	$(DEBUG) -l

CFLAGS =	-a $(DEBUG) -D_TRH=0

TSOURCES =	testtcp.c worms.c
LSOURCES =	tgetent.c tgetflag.c tgetnum.c tgetstr.c tgoto.c tputs.c \
		_tcpfind.c fgetlr.c

HEADERS =	_termcap.h termcap.h

TOBJECTS =	testtcp.o worms.o tset.o
LOBJECTS =	tgetent.o tgetflag.o tgetnum.o tgetstr.o tgoto.o tputs.o \
		_tcpfind.o fgetlr.o

LIBDIR =	$(LIB)

#TRHLIB =	$(LIB)trh.lib	# for error()

termcap.lib:	$(LOBJECTS)
	@echo Now put all the object files into library "libtermcap.a"
	tlink -j $(DEBUG) -o=termcap.lib $(LOBJECTS)

$(LOBJECTS) tset.o:	_termcap.h

install: termcap.lib
	cp termcap.lib $(LIBDIR)

all:	termcap.lib tests

tests: 	testtcp.ttp worms.prg

worms.prg:	worms.o
	linkgem worms termcap.lib $(DEBUG) -s=20000

testtcp.ttp: testtcp.o
	linkgem testtcp termcap.lib $(DEBUG)

#tset:	tset.o
#	$(CC) -o tset tset.o -ltermcap $(TRHLIB)
	
#
#	Clean up the directory.
#
clean:
	rm -f *.o testtcp.ttp worms.prg termcap.lib

tags:	$(HEADERS) $(LSOURCES) tset.c $(TSOURCES) 
	ctags -t $(HEADERS) $(LSOURCES) tset.c $(TSOURCES)
