/*+++*
 *  RCS		$Id: tinfocap.h,v 1.1 1993/07/05 00:20:48 tom Exp tom $
 *  title:	tinfocap.h
 *  abstract:	definitions for termcap emulation routines based on terminfo.
 *  author:	T.R.Hageman, Groningen, The Netherlands.
 *  created:	July 1993, extracted from tinfocap.c
 *  modified:	(see RCS Log at end of header)
 *  description:
 *	This contains the table definitions for the portable termcap-style
 *	access routines to the terminfo database.
 *
#if NOTYET
 *	If the compile-time option TI_TERMCAP is #defined, the names of
 *	the termcap routines in this file are prepended with "ti_", and the
 *	standard termcap is also accessed.  The value of TI_TERMCAP determines
 *	whether termcap or terminfo is accessed first:
 *	  TI_TERMCAP > 0	first TERMCAP, then TERMINFO
 *	  TI_TERMCAP < 0	first TERMINFO, then TERMCAP
 *	  TI_TERMCAP = 0	determined at runtime by value of `_ti_termcap'.
#endif
 *
 *  WARNING:
 *	This assumes the conventional SYSV/AT&T terminfo implementation,
 *	i.e., all capability names are #defined in <term.h>,
 *	and are stored in a single structure, referenced by CUR.
 *
 *	The SCO implementation using CURB, CURN, and CURS for separate
 *	flag-, numeric-, and string capability tables is also supported.
 *	(except for printer-describing capabilities)
 *
 *	This implementation works by redefining CUR (+ CURB, CURN, CURS)
 *	to a constant pointer at table definition time, so do NOT count
 *	on CUR (or *any* variable defined by <term.h>, for that matter)
 *	being correct after that.
 *
 *	For other implementations this may not work.
 *
 *  usage:
 *	#include this file once as the table's contents for each table
 *	you want to define, with appropriate definitions of
 *
 *		ENTRY(i,d, ti_name, pad_parm, ti_variable)
 *
 *	where the two characters `i',`d' form the entry's termcap name,
 *	the string `ti_name' is its terminfo name,
 *	the number `pad_parm' encodes the number of parameters expected
 *	with flags indicating padding (PAD), padding with multiple lines
 *	(PADM) (whicm implies PAD, incidentally) and whether it is a key
 *	sequence (KEY),
 *	the identifier `ti_variable' is the variable's name as defined by
 *	terminfo;
 *	and
 *
 *		TABLE
 *
 *	which should be one of
 *
 *		FLAGS, NUMBERS, or STRINGS.
 *
 *	If TABLE is undefined this can be used as an ordinary header file.
 *
 *	#include <term.h> before this one.
 *---*/


/* Build an integer key from a two-character termcap name. */

#define	TC_ID(i,d)	((i)<<8|(d))


/* Special flags OR-ed in with the number of parameters.
   (this allows for maximum of 31 parameters which is plenty.) */

#define PAD	0x20
#define PADM	(PAD|0x40)
#define KEY	0x80

#define PARMASK	(PAD-1)


/* Special handling for SCO terminfo, which uses separate tables
   for its non-standard boolean, numeric and string capabilities.
   We have to know about some internals of it, unfortunately.
   The ?_OFFSET defines are used to distinguish between the tables,
   and are chosen such that (B_OFFSET < N_OFFSET < S_OFFSET), and
   that string capabilities have plenty of room for extension. */

#ifdef CURB
#   define _B_FIRST_FIELD	((char *) &CURB _auto_left_margin)
#   define _B_OFFSET		0x2000
#   define _B_OFFSET_TO_TI(offset)	\
	(offset) >= _B_OFFSET ? _B_FIRST_FIELD + ((offset) - _B_OFFSET) :  
#else
#   define _B_OFFSET_TO_TI(offset)	/* dummy */
#endif

#ifdef CURN
#   define _N_FIRST_FIELD	((char *) &CURN _columns)
#   define _N_OFFSET		0x3000
#   define _N_OFFSET_TO_TI(offset)	\
	(offset) >= _N_OFFSET ? _N_FIRST_FIELD + ((offset) - _N_OFFSET) :  
#else
#   define _N_OFFSET_TO_TI(offset)	/* dummy */
#endif

#ifdef CURS
#   define _S_FIRST_FIELD	((char *) &CURS strs)
#   define _S_OFFSET		0x4000
#   define _S_OFFSET_TO_TI(offset)	\
	(offset) >= _S_OFFSET ? _S_FIRST_FIELD + ((offset) - _S_OFFSET) :  
#else
#   define _S_OFFSET_TO_TI(offset)	/* dummy */
#endif


/* First field defined in struct term.  This is used in offset calculations. */

#if (_B_OFFSET && _N_OFFSET && _S_OFFSET)
#   define _FIRST_FIELD		((char *) &CUR _b1)	/* Just to be sure... */
#else
#   define _FIRST_FIELD		((char *) &auto_left_margin)
#endif


/* Convert the address of `field' (in struct term) to an offset
   (relative to the first field in struct term.)
   This is independent of the actual table it is in (for SCO) due to the
   redefinition tricks below. */

#define TI_TO_OFFSET(field)	((char *)&(field) - _FIRST_FIELD)


/* Convert an `offset' (calculated by the TI_TO_OFFSET macro above) back
   to the address of a field, cast to (char *). */

#define OFFSET_TO_TI(offset)	(_S_OFFSET_TO_TI(offset) \
				 _N_OFFSET_TO_TI(offset) \
				 _B_OFFSET_TO_TI(offset) \
				 (offset) + _FIRST_FIELD)


#ifdef TABLE

#define FLAGS	1
#define NUMBERS	2
#define STRINGS	3

/* Sanity checks. */

#if (TABLE-0 < FLAGS || STRINGS < TABLE-0)
    #error "TABLE should be #defined as one of `FLAGS', `NUMBERS' or `STRINGS'."
#endif


/* ENTRY(i,d, terminfo_name, pad_parm, terminfo_field)

   Define an entry in a capabilities table.

   "Standard" terminfo entries are defined using ENTRY.
   "Extended" entries are defined using EXTRY for easy recognition,
   and are surrounded with:
	#if ALL_ENTRIES || defined (<variable-name>)
	...
	#endif
   so these are only compiled in when defined in <term.h>, or when the
   compile-time option ALL_ENTRIES is defined (this should only be
   defined when your definition of ENTRY does not reference `ti_variable'
   or else you'll get a host of compilation errors...)

   The tables should be lexically ordered by termcap id. (i,d) */

#ifndef ENTRY
    #error "ENTRY(i,d, terminfo_name, pad_flags, terminfo_field) is undefined."
#endif


/* EXTRY is normally used as an alias for ENTRY, to easily recognize
   various extensions to the (original) termcap capabilities.
   Define it here as such if it isn't already defined. */

#ifndef EXTRY
#   define EXTRY	ENTRY
#endif


/* Redefine CUR to a constant pointer.  (This is to avoid problems with
   compilers that cannot evaluate the pointer difference expression at
   compile-time.  It is ABSOLUTELY ESSENTIAL for SCO where terminfo variables
   are scattered over several structures.) */

#undef CUR
#define CUR	((struct term *)0)->


#if (TABLE-0 == FLAGS)		/* ======== Boolean capabilities. ======== */

#   ifdef CURB
#	undef  CURB
#	define CURB	((struct _bool_struct *)_B_OFFSET)->
#   endif

#if ALL_ENTRIES || defined(prtr_silent)
EXTRY('5','i', "mc5i",	0,	prtr_silent )		/* printer will not echo on screen */
#endif
#if ALL_ENTRIES || defined(hard_cursor)
EXTRY('H','C', "chts",	0,	hard_cursor )		/* Cursor is hard to see */
#endif
#if ALL_ENTRIES || defined(non_dest_scroll_region)
EXTRY('N','D', "ndscr",	0,	non_dest_scroll_region )	/* (SCO) Non-destructive scroll region */
#endif
#if ALL_ENTRIES || defined(no_pad_char)
EXTRY('N','P', "npc",	0,	no_pad_char )		/* (SCO) Pad character doesn't exist */
#endif
#if ALL_ENTRIES || defined(non_rev_rmcup)
EXTRY('N','R', "nrrmc",	0,	non_rev_rmcup )		/* (SCO) smcup does not reverse rmcup */
#endif
ENTRY('a','m', "am",	0,	auto_right_margin )	/* Terminal has automatic margins */
#if ALL_ENTRIES || defined(back_color_erase)
EXTRY('b','e', "bce",	0,	back_color_erase )	/* (SCO) Screen erased with background color */
#endif
ENTRY('b','w', "bw",	0,	auto_left_margin )	/* cub1 wraps from column 0 to last column */
#if ALL_ENTRIES || defined(can_change)
EXTRY('c','c', "ccc",	0,	can_change )		/* (SCO) Terminal can re-define existing color */
#endif
ENTRY('d','a', "da",	0,	memory_above )		/* Display may be retained above the screen */
ENTRY('d','b', "db",	0,	memory_below )		/* Display may be retained below the screen */
ENTRY('e','o', "eo",	0,	erase_overstrike )	/* Can erase overstrikes with a blank */
ENTRY('e','s', "eslok",	0,	status_line_esc_ok )	/* Escape can be used on the status line */
ENTRY('g','n', "gn",	0,	generic_type )		/* Generic line type (e.g.,, dialup, switch) */
ENTRY('h','c', "hc",	0,	hard_copy )		/* Hardcopy terminal */
#if ALL_ENTRIES || defined(hue_lightness_saturation)
EXTRY('h','l', "hls",	0,	hue_lightness_saturation )	/* (SCO) Hue-lightness-saturation color scheme */
#endif
ENTRY('h','s', "hs",	0,	has_status_line )	/* Has extra "status line" */
ENTRY('h','z', "hz",	0,	tilde_glitch )		/* Hazeltine; can not print ~'s */
ENTRY('i','n', "in",	0,	insert_null_glitch )	/* Insert mode distinguishes nulls */
ENTRY('k','m', "km",	0,	has_meta_key )		/* Has a meta key (shift, sets parity bit) */
ENTRY('m','i', "mir",	0,	move_insert_mode )	/* Safe to move while in insert mode */
ENTRY('m','s', "msgr",	0,	move_standout_mode )	/* Safe to move in standout modes */
#if ALL_ENTRIES || defined(needs_xon_xoff)
EXTRY('n','x', "nxon",	0,	needs_xon_xoff )	/* XOn/Xoff required */
#endif
ENTRY('o','s', "os",	0,	over_strike )		/* Terminal overstrikes */
ENTRY('u','l', "ul",	0,	transparent_underline )	/* underline character overstrikes */
#ifndef no_esc_ctlc
#   define no_esc_ctlc	beehive_glitch
#endif
ENTRY('x','b', "xsb",	0,	no_esc_ctlc )		/* Beehive (f1=escape, f2=ctrl C) */
ENTRY('x','n', "xenl",	0,	eat_newline_glitch )	/* newline ignored after 80 cols (Concept) */
ENTRY('x','o', "xon",	0,	xon_xoff )		/* Terminal uses xon/xoff handshaking */
ENTRY('x','s', "xhp",	0,	ceol_standout_glitch )	/* Standout not erased by overwriting (hp) */
#ifndef dest_tabs_magic_smso
#   define dest_tabs_magic_smso	teleray_glitch
#endif
ENTRY('x','t', "xt",	0,	dest_tabs_magic_smso )	/* Tabs ruin, magic so char (Teleray 1061) */

#endif /* FLAGS */



#if (TABLE-0 == NUMBERS)	/* ======== Numeric capabilities. ======== */

#ifdef CURN
#   undef  CURN
#   define CURN	((struct _num_struct *)_N_OFFSET)->
#endif

#if ALL_ENTRIES || defined(max_colors)
EXTRY('C','o', "colors",	0,	max_colors )	/* (SCO) Maximum number of colors on the screen */
#endif
#if ALL_ENTRIES || defined(maximum_windows)
EXTRY('M','W', "wnum",	0,	maximum_windows )	/* (SCO) Maximum number of definable windows */
#endif
#if ALL_ENTRIES || defined(no_color_video)
EXTRY('N','C', "ncv",	0,	no_color_video )	/* (SCO) Video attributes that can't be used with colors */
#endif
#if ALL_ENTRIES || defined(num_labels)
EXTRY('N','l', "nlab",	0,	num_labels )		/* (SCO) Number of labels at screen (start at 1) */
#endif
ENTRY('c','o', "cols",	0,	columns )		/* Number of columns in a line */
ENTRY('i','t', "it",	0,	init_tabs )		/* Tabs initially every # spaces */
ENTRY('l','i', "lines",	0,	lines )			/* Number of lines on screen or page */
#if ALL_ENTRIES || defined(label_height)
EXTRY('l','h', "lh",	0,	label_height )		/* (SCO) Number of rows in each label */
#endif
ENTRY('l','m', "lm",	0,	lines_of_memory )	/* Lines of memory if > lines.  0 means varies */
#if ALL_ENTRIES || defined(label_width)
EXTRY('l','w', "lw",	0,	label_width )		/* Number of columns in each label */
#endif
#if ALL_ENTRIES || defined(max_attributes)
EXTRY('m','a', "ma",	0,	max_attributes )	/* (SCO) Maximum combined video attributes terminal can display */
#endif
#if ALL_ENTRIES || defined(max_pairs)
EXTRY('p','a', "pairs",	0,	max_pairs )		/* (SCO) Maximum number of color-pairs on the screen */
#endif
ENTRY('p','b', "pb",	0,	padding_baud_rate )	/* Lowest baud where cr/nl padding is needed */
ENTRY('s','g', "xmc",	0,	magic_cookie_glitch )	/* Number of blank chars left by smso or rmso */
ENTRY('v','t', "vt",	0,	virtual_terminal )	/* Virtual terminal number (UNIX system) */
ENTRY('w','s', "wsl",	0,	width_status_line )	/* Number of columns in status line */

#endif /* NUMBERS */


#if (TABLE-0 == STRINGS)	/* ======== String capabilities. ======== */

#ifdef CURS
#   undef  CURS
#   define CURS	((struct _str_struct *)_S_OFFSET)->
#endif

#if ALL_ENTRIES || defined(key_ssave)
EXTRY('!','1', "kSAV",	KEY,	key_ssave )
#endif
#if ALL_ENTRIES || defined(key_ssuspend)
EXTRY('!','2', "kSPD",	KEY,	key_ssuspend )
#endif
#if ALL_ENTRIES || defined(key_sundo)
EXTRY('!','3', "kUND",	KEY,	key_sundo )
#endif
#if ALL_ENTRIES || defined(key_shelp)
EXTRY('#','1', "kHLP",	KEY,	key_shelp )
#endif
#if ALL_ENTRIES || defined(key_shome)
EXTRY('#','2', "kHOM",	KEY,	key_shome )
#endif
#if ALL_ENTRIES || defined(key_sic)
EXTRY('#','3', "kIC",	KEY,	key_sic )
#endif
#if ALL_ENTRIES || defined(key_sleft)
EXTRY('#','4', "kLFT",	KEY,	key_sleft )
#endif
#if ALL_ENTRIES || defined(key_redo)
EXTRY('%','0', "krdo",	KEY,	key_redo )
#endif
#if ALL_ENTRIES || defined(key_help)
EXTRY('%','1', "khlp",	KEY,	key_help )
#endif
#if ALL_ENTRIES || defined(key_mark)
EXTRY('%','2', "kmrk",	KEY,	key_mark )
#endif
#if ALL_ENTRIES || defined(key_message)
EXTRY('%','3', "kmsg",	KEY,	key_message )
#endif
#if ALL_ENTRIES || defined(key_move)
EXTRY('%','4', "kmov",	KEY,	key_move )
#endif
#if ALL_ENTRIES || defined(key_next)
EXTRY('%','5', "knxt",	KEY,	key_next )
#endif
#if ALL_ENTRIES || defined(key_open)
EXTRY('%','6', "kopn",	KEY,	key_open )
#endif
#if ALL_ENTRIES || defined(key_options)
EXTRY('%','7', "kopt",	KEY,	key_options )
#endif
#if ALL_ENTRIES || defined(key_previous)
EXTRY('%','8', "kprv",	KEY,	key_previous )
#endif
#if ALL_ENTRIES || defined(key_print)
EXTRY('%','9', "kprt",	KEY,	key_print )
#endif
#if ALL_ENTRIES || defined(key_smessage)
EXTRY('%','a', "kMSG",	KEY,	key_smessage )
#endif
#if ALL_ENTRIES || defined(key_smove)
EXTRY('%','b', "kMOV",	KEY,	key_smove )
#endif
#if ALL_ENTRIES || defined(key_snext)
EXTRY('%','c', "kNXT",	KEY,	key_snext )
#endif
#if ALL_ENTRIES || defined(key_soptions)
EXTRY('%','d', "kOPT",	KEY,	key_soptions )
#endif
#if ALL_ENTRIES || defined(key_sprevious)
EXTRY('%','e', "kPRV",	KEY,	key_sprevious )
#endif
#if ALL_ENTRIES || defined(key_sprint)
EXTRY('%','f', "kPRT",	KEY,	key_sprint )
#endif
#if ALL_ENTRIES || defined(key_sredo)
EXTRY('%','g', "kRDO",	KEY,	key_sredo )
#endif
#if ALL_ENTRIES || defined(key_sreplace)
EXTRY('%','h', "kRPL",	KEY,	key_sreplace )
#endif
#if ALL_ENTRIES || defined(key_sright)
EXTRY('%','i', "kRIT",	KEY,	key_sright )
#endif
#ifndef key_sresume
#   ifdef key_srsume
#	define key_sresume	key_srsume		/* (SCO) typo? */
#   endif
#endif
#if ALL_ENTRIES || defined(key_sresume)
EXTRY('%','j', "kRES",	KEY,	key_sresume )
#endif
#if ALL_ENTRIES || defined(key_scancel)
EXTRY('&','0', "kCAN",	KEY,	key_scancel )
#endif
#if ALL_ENTRIES || defined(key_reference)
EXTRY('&','1', "kref",	KEY,	key_reference )
#endif
#if ALL_ENTRIES || defined(key_refresh)
EXTRY('&','2', "krfr",	KEY,	key_refresh )
#endif
#if ALL_ENTRIES || defined(key_replace)
EXTRY('&','3', "krpl",	KEY,	key_replace )
#endif
#if ALL_ENTRIES || defined(key_restart)
EXTRY('&','4', "krst",	KEY,	key_restart )
#endif
#if ALL_ENTRIES || defined(key_resume)
EXTRY('&','5', "kres",	KEY,	key_resume )
#endif
#if ALL_ENTRIES || defined(key_save)
EXTRY('&','6', "ksav",	KEY,	key_save )
#endif
#if ALL_ENTRIES || defined(key_suspend)
EXTRY('&','7', "kspd",	KEY,	key_suspend )
#endif
#if ALL_ENTRIES || defined(key_undo)
EXTRY('&','8', "kund",	KEY,	key_undo )
#endif
#if ALL_ENTRIES || defined(key_sbeg)
EXTRY('&','9', "kBEG",	KEY,	key_sbeg )
#endif
#if ALL_ENTRIES || defined(key_sfind)
EXTRY('*','0', "kFND",	KEY,	key_sfind )
#endif
#if ALL_ENTRIES || defined(key_scommand)
EXTRY('*','1', "kCMD",	KEY,	key_scommand )
#endif
#if ALL_ENTRIES || defined(key_scopy)
EXTRY('*','2', "kCPY",	KEY,	key_scopy )
#endif
#if ALL_ENTRIES || defined(key_screate)
EXTRY('*','3', "kCRT",	KEY,	key_screate )
#endif
#if ALL_ENTRIES || defined(key_sdc)
EXTRY('*','4', "kDC",	KEY,	key_sdc )
#endif
#if ALL_ENTRIES || defined(key_sdl)
EXTRY('*','5', "kDL",	KEY,	key_sdl )
#endif
#if ALL_ENTRIES || defined(key_select)
EXTRY('*','6', "kslt",	KEY,	key_select )
#endif
#if ALL_ENTRIES || defined(key_send)
EXTRY('*','7', "kEND",	KEY,	key_send )
#endif
#if ALL_ENTRIES || defined(key_seol)
EXTRY('*','8', "kEOL",	KEY,	key_seol )
#endif
#if ALL_ENTRIES || defined(key_sexit)
EXTRY('*','9', "kEXT",	KEY,	key_sexit )
#endif
#if ALL_ENTRIES || defined(key_find)
EXTRY('@','0', "kfnd",	KEY,	key_find )
#endif
#if ALL_ENTRIES || defined(key_beg)
EXTRY('@','1', "kbeg",	KEY,	key_beg )
#endif
#if ALL_ENTRIES || defined(key_cancel)
EXTRY('@','2', "kcan",	KEY,	key_cancel )
#endif
#if ALL_ENTRIES || defined(key_close)
EXTRY('@','3', "kclo",	KEY,	key_close )
#endif
#if ALL_ENTRIES || defined(key_command)
EXTRY('@','4', "kcmd",	KEY,	key_command )
#endif
#if ALL_ENTRIES || defined(key_copy)
EXTRY('@','5', "kcpy",	KEY,	key_copy )
#endif
#if ALL_ENTRIES || defined(key_create)
EXTRY('@','6', "kcrt",	KEY,	key_create )
#endif
#if ALL_ENTRIES || defined(key_end)
EXTRY('@','7', "kend",	KEY,	key_end )
#endif
#if ALL_ENTRIES || defined(key_enter)
EXTRY('@','8', "kent",	KEY,	key_enter )
#endif
#if ALL_ENTRIES || defined(key_exit)
EXTRY('@','9', "kext",	KEY,	key_exit )
#endif
ENTRY('A','L', "il",	PADM|1,	parm_insert_line )	/* Add #1 new blank lines (PG*) */
#if ALL_ENTRIES || defined(box_attr_1)
EXTRY('B','x', "batt1",	0,	box_attr_1 )		/* (AIX) Attributes for box_chars_1 */
#endif
#if ALL_ENTRIES || defined(box_attr_2)
EXTRY('B','y', "batt2",	0,	box_attr_2 )		/* (AIX) Attributes for box_chars_2 */
#endif
ENTRY('C','C', "cmdch",	0,	command_character )	/* Terminal settable cmd char in prototype */
ENTRY('C','M', "mrcup",	PAD|2,	cursor_mem_address )	/* Memory relative cursor addressing */
#if ALL_ENTRIES || defined(create_window)
EXTRY('C','W', "cwin",	5,	create_window )		/* (SCO) Window #1 to #2..#5 */
#endif
ENTRY('D','C', "dch",	PADM|1,	parm_dch )		/* Delete #1 chars (PG*) */
#if ALL_ENTRIES || defined(dial_phone)
EXTRY('D','I', "dial",	1,	dial_phone )		/* (SCO) Dial phone number #1.*/
#endif
#if ALL_ENTRIES || defined(display_clock)
EXTRY('D','K', "dclk",	0,	display_clock )		/* (SCO) Display time-of-day clock */
#endif
ENTRY('D','L', "dl",	PADM|1,	parm_delete_line )	/* Delete #1 lines (PG*) */
ENTRY('D','O', "cud",	PADM|1,	parm_down_cursor )	/* Move cursor down #1 lines (PG*) */
#if ALL_ENTRIES || defined(key_f11)
EXTRY('F','1', "kf11",	KEY,	key_f11 )		/* Sent by function key f11.*/
#endif
#if ALL_ENTRIES || defined(key_f12)
EXTRY('F','2', "kf12",	KEY,	key_f12 )		/* Sent by function key f12.*/
#endif
#if ALL_ENTRIES || defined(key_f13)
EXTRY('F','3', "kf13",	KEY,	key_f13 )		/* Sent by function key f13.*/
#endif
#if ALL_ENTRIES || defined(key_f14)
EXTRY('F','4', "kf14",	KEY,	key_f14 )		/* Sent by function key f14.*/
#endif
#if ALL_ENTRIES || defined(key_f15)
EXTRY('F','5', "kf15",	KEY,	key_f15 )		/* Sent by function key f15.*/
#endif
#if ALL_ENTRIES || defined(key_f16)
EXTRY('F','6', "kf16",	KEY,	key_f16 )		/* Sent by function key f16.*/
#endif
#if ALL_ENTRIES || defined(key_f17)
EXTRY('F','7', "kf17",	KEY,	key_f17 )		/* Sent by function key f17.*/
#endif
#if ALL_ENTRIES || defined(key_f18)
EXTRY('F','8', "kf18",	KEY,	key_f18 )		/* Sent by function key f18.*/
#endif
#if ALL_ENTRIES || defined(key_f19)
EXTRY('F','9', "kf19",	KEY,	key_f19 )		/* Sent by function key f19.*/
#endif
#if ALL_ENTRIES || defined(key_f20)
EXTRY('F','A', "kf20",	KEY,	key_f20 )		/* Sent by function key f20.*/
#endif
#if ALL_ENTRIES || defined(key_f21)
EXTRY('F','B', "kf21",	KEY,	key_f21 )		/* Sent by function key f21.*/
#endif
#if ALL_ENTRIES || defined(key_f22)
EXTRY('F','C', "kf22",	KEY,	key_f22 )		/* Sent by function key f22.*/
#endif
#if ALL_ENTRIES || defined(key_f23)
EXTRY('F','D', "kf23",	KEY,	key_f23 )		/* Sent by function key f23.*/
#endif
#if ALL_ENTRIES || defined(key_f24)
EXTRY('F','E', "kf24",	KEY,	key_f24 )		/* Sent by function key f24.*/
#endif
#if ALL_ENTRIES || defined(key_f25)
EXTRY('F','F', "kf25",	KEY,	key_f25 )		/* Sent by function key f25.*/
#endif
#if ALL_ENTRIES || defined(key_f26)
EXTRY('F','G', "kf26",	KEY,	key_f26 )		/* Sent by function key f26.*/
#endif
#if ALL_ENTRIES || defined(key_f27)
EXTRY('F','H', "kf27",	KEY,	key_f27 )		/* Sent by function key f27.*/
#endif
#if ALL_ENTRIES || defined(key_f28)
EXTRY('F','I', "kf28",	KEY,	key_f28 )		/* Sent by function key f28.*/
#endif
#if ALL_ENTRIES || defined(key_f29)
EXTRY('F','J', "kf29",	KEY,	key_f29 )		/* Sent by function key f29.*/
#endif
#if ALL_ENTRIES || defined(key_f30)
EXTRY('F','K', "kf30",	KEY,	key_f30 )		/* Sent by function key f30.*/
#endif
#if ALL_ENTRIES || defined(key_f31)
EXTRY('F','L', "kf31",	KEY,	key_f31 )		/* Sent by function key f31.*/
#endif
#if ALL_ENTRIES || defined(key_f32)
EXTRY('F','M', "kf32",	KEY,	key_f32 )		/* Sent by function key f32.*/
#endif
#if ALL_ENTRIES || defined(key_f33)
EXTRY('F','N', "kf33",	KEY,	key_f33 )		/* Sent by function key f33.*/
#endif
#if ALL_ENTRIES || defined(key_f34)
EXTRY('F','O', "kf34",	KEY,	key_f34 )		/* Sent by function key f34.*/
#endif
#if ALL_ENTRIES || defined(key_f35)
EXTRY('F','P', "kf35",	KEY,	key_f35 )		/* Sent by function key f35.*/
#endif
#if ALL_ENTRIES || defined(key_f36)
EXTRY('F','Q', "kf36",	KEY,	key_f36 )		/* Sent by function key f36.*/
#endif
#if ALL_ENTRIES || defined(key_f37)
EXTRY('F','R', "kf37",	KEY,	key_f37 )		/* Sent by function key f37.*/
#endif
#if ALL_ENTRIES || defined(key_f38)
EXTRY('F','S', "kf38",	KEY,	key_f38 )		/* Sent by function key f38.*/
#endif
#if ALL_ENTRIES || defined(key_f39)
EXTRY('F','T', "kf39",	KEY,	key_f39 )		/* Sent by function key f39.*/
#endif
#if ALL_ENTRIES || defined(key_f40)
EXTRY('F','U', "kf40",	KEY,	key_f40 )		/* Sent by function key f40.*/
#endif
#if ALL_ENTRIES || defined(key_f41)
EXTRY('F','V', "kf41",	KEY,	key_f41 )		/* Sent by function key f41.*/
#endif
#if ALL_ENTRIES || defined(key_f42)
EXTRY('F','W', "kf42",	KEY,	key_f42 )		/* Sent by function key f42.*/
#endif
#if ALL_ENTRIES || defined(key_f43)
EXTRY('F','X', "kf43",	KEY,	key_f43 )		/* Sent by function key f43.*/
#endif
#if ALL_ENTRIES || defined(key_f44)
EXTRY('F','Y', "kf44",	KEY,	key_f44 )		/* Sent by function key f44.*/
#endif
#if ALL_ENTRIES || defined(key_f45)
EXTRY('F','Z', "kf45",	KEY,	key_f45 )		/* Sent by function key f45.*/
#endif
#if ALL_ENTRIES || defined(key_f46)
EXTRY('F','a', "kf46",	KEY,	key_f46 )		/* Sent by function key f46.*/
#endif
#if ALL_ENTRIES || defined(key_f47)
EXTRY('F','b', "kf47",	KEY,	key_f47 )		/* Sent by function key f47.*/
#endif
#if ALL_ENTRIES || defined(key_f48)
EXTRY('F','c', "kf48",	KEY,	key_f48 )		/* Sent by function key f48.*/
#endif
#if ALL_ENTRIES || defined(key_f49)
EXTRY('F','d', "kf49",	KEY,	key_f49 )		/* Sent by function key f49.*/
#endif
#if ALL_ENTRIES || defined(key_f50)
EXTRY('F','e', "kf50",	KEY,	key_f50 )		/* Sent by function key f50.*/
#endif
#if ALL_ENTRIES || defined(key_f51)
EXTRY('F','f', "kf51",	KEY,	key_f51 )		/* Sent by function key f51.*/
#endif
#if ALL_ENTRIES || defined(key_f52)
EXTRY('F','g', "kf52",	KEY,	key_f52 )		/* Sent by function key f52.*/
#endif
#if ALL_ENTRIES || defined(key_f53)
EXTRY('F','h', "kf53",	KEY,	key_f53 )		/* Sent by function key f53.*/
#endif
#if ALL_ENTRIES || defined(key_f54)
EXTRY('F','i', "kf54",	KEY,	key_f54 )		/* Sent by function key f54.*/
#endif
#if ALL_ENTRIES || defined(key_f55)
EXTRY('F','j', "kf55",	KEY,	key_f55 )		/* Sent by function key f55.*/
#endif
#if ALL_ENTRIES || defined(key_f56)
EXTRY('F','k', "kf56",	KEY,	key_f56 )		/* Sent by function key f56.*/
#endif
#if ALL_ENTRIES || defined(key_f57)
EXTRY('F','l', "kf57",	KEY,	key_f57 )		/* Sent by function key f57.*/
#endif
#if ALL_ENTRIES || defined(key_f58)
EXTRY('F','m', "kf58",	KEY,	key_f58 )		/* Sent by function key f58.*/
#endif
#if ALL_ENTRIES || defined(key_f59)
EXTRY('F','n', "kf59",	KEY,	key_f59 )		/* Sent by function key f59.*/
#endif
#if ALL_ENTRIES || defined(key_f60)
EXTRY('F','o', "kf60",	KEY,	key_f60 )		/* Sent by function key f60.*/
#endif
#if ALL_ENTRIES || defined(key_f61)
EXTRY('F','p', "kf61",	KEY,	key_f61 )		/* Sent by function key f61.*/
#endif
#if ALL_ENTRIES || defined(key_f62)
EXTRY('F','q', "kf62",	KEY,	key_f62 )		/* Sent by function key f62.*/
#endif
#if ALL_ENTRIES || defined(key_f63)
EXTRY('F','r', "kf63",	KEY,	key_f63 )		/* Sent by function key f63.*/
#endif
#if ALL_ENTRIES || defined(hangup)
EXTRY('H','U', "hup",	0,	hangup )		/* (SCO) Hangup phone */
#endif
ENTRY('I','C', "ich",	PADM|1,	parm_ich )		/* Insert #1 blank chars (PG*) */
#if ALL_ENTRIES || defined(initialize_color)
EXTRY('I','c', "initc",	4,	initialize_color )	/* (SCO) Initialize color #1 to #2,#3,#4.*/
#endif
#if ALL_ENTRIES || defined(initialize_pair)
EXTRY('I','p', "initp",	7,	initialize_pair )	/* (SCO) Initialize color pair to fg #2..#4, bg #5..#7.*/
#endif
ENTRY('K','1', "ka1",	KEY,	key_a1 )		/* Upper left of keypad */
ENTRY('K','2', "kb2",	KEY,	key_b2 )		/* Center of keypad */
ENTRY('K','3', "ka3",	KEY,	key_a3 )		/* Upper right of keypad */
ENTRY('K','4', "kc1",	KEY,	key_c1 )		/* Lower left of keypad */
ENTRY('K','5', "kc3",	KEY,	key_c3 )		/* Lower right of keypad */
#if ALL_ENTRIES || defined(key_smap_out6)
EXTRY('K','R', "kmpt6",	KEY,	key_smap_out6 )		/* (AIX) output for mapped key 6*/
#endif
#if ALL_ENTRIES || defined(key_smap_out7)
EXTRY('K','S', "kmpt7",	KEY,	key_smap_out7 )		/* (AIX) output for mapped key 7*/
#endif
#if ALL_ENTRIES || defined(key_smap_out8)
EXTRY('K','T', "kmpt8",	KEY,	key_smap_out8 )		/* (AIX) output for mapped key 8*/
#endif
#if ALL_ENTRIES || defined(key_smap_out9)
EXTRY('K','U', "kmpt9",	KEY,	key_smap_out9 )		/* (AIX) output for mapped key 9*/
#endif
#if ALL_ENTRIES || defined(key_smap_out1)
EXTRY('K','V', "kmpt1",	KEY,	key_smap_out1 )		/* (AIX) output for mapped key 1*/
#endif
#if ALL_ENTRIES || defined(key_smap_out2)
EXTRY('K','W', "kmpt2",	KEY,	key_smap_out2 )		/* (AIX) output for mapped key 2*/
#endif
#if ALL_ENTRIES || defined(key_smap_out3)
EXTRY('K','X', "kmpt3",	KEY,	key_smap_out3 )		/* (AIX) output for mapped key 3*/
#endif
#if ALL_ENTRIES || defined(key_smap_out4)
EXTRY('K','Y', "kmpt4",	KEY,	key_smap_out4 )		/* (AIX) output for mapped key 4*/
#endif
#if ALL_ENTRIES || defined(key_smap_out5)
EXTRY('K','Z', "kmpt5",	KEY,	key_smap_out5 )		/* (AIX) output for mapped key 5*/
#endif
#if ALL_ENTRIES || defined(key_smap_in6)
EXTRY('K','r', "kmpf6",	KEY,	key_smap_in6 )		/* (AIX) input for special mapped key 6*/
#endif
#if ALL_ENTRIES || defined(key_smap_in7)
EXTRY('K','s', "kmpf7",	KEY,	key_smap_in7 )		/* (AIX) input for special mapped key 7*/
#endif
#if ALL_ENTRIES || defined(key_smap_in8)
EXTRY('K','t', "kmpf8",	KEY,	key_smap_in8 )		/* (AIX) input for special mapped key 8*/
#endif
#if ALL_ENTRIES || defined(key_smap_in9)
EXTRY('K','u', "kmpf9",	KEY,	key_smap_in9 )		/* (AIX) input for special mapped key 9*/
#endif
#if ALL_ENTRIES || defined(key_smap_in1)
EXTRY('K','v', "kmpf1",	KEY,	key_smap_in1 )		/* (AIX) input for special mapped key 1*/
#endif
#if ALL_ENTRIES || defined(key_smap_in2)
EXTRY('K','w', "kmpf2",	KEY,	key_smap_in2 )		/* (AIX) input for special mapped key 2*/
#endif
#if ALL_ENTRIES || defined(key_smap_in3)
EXTRY('K','x', "kmpf3",	KEY,	key_smap_in3 )		/* (AIX) input for special mapped key 3*/
#endif
#if ALL_ENTRIES || defined(key_smap_in4)
EXTRY('K','y', "kmpf4",	KEY,	key_smap_in4 )		/* (AIX) input for special mapped key 4*/
#endif
#if ALL_ENTRIES || defined(key_smap_in5)
EXTRY('K','z', "kmpf5",	KEY,	key_smap_in5 )		/* (AIX) input for special mapped key 5*/
#endif
ENTRY('L','E', "cub",	PAD|1,	parm_left_cursor )	/* Move cursor left #1 spaces (PG) */
#if ALL_ENTRIES || defined(label_off)
EXTRY('L','F', "rmln",	0,	label_off )		/* Turn off soft labels */
#endif
#if ALL_ENTRIES || defined(label_on)
EXTRY('L','O', "smln",	0,	label_on )		/* Turn on soft labels */
#endif
#if ALL_ENTRIES || defined(label_format)
EXTRY('L','f', "fln",	0/*?*/,	label_format )		/* (SCO) Label format (???) */
#endif
#if ALL_ENTRIES || defined(clear_margins)
EXTRY('M','C', "mgc",	0,	clear_margins )		/* Clear all margins (top, bottom and sides) */
#endif
#if ALL_ENTRIES || defined(set_left_margin)
EXTRY('M','L', "smgl",	0,	set_left_margin )	/* Set left margin at current column */
#endif
#if ALL_ENTRIES || defined(set_right_margin)
EXTRY('M','R', "smgr",	0,	set_right_margin )	/* Set right margin at current column */
#endif
#if ALL_ENTRIES || defined(fixed_pause)
EXTRY('P','A', "pause",	0,	fixed_pause )		/* (SCO) Pause for 2-3 seconds */
#endif
#if ALL_ENTRIES || defined(pulse)
EXTRY('P','U', "pulse",	0,	pulse )			/* (SCO) Select pulse code dialing */
#endif
#if ALL_ENTRIES || defined(quick_dial)
EXTRY('Q','D', "qdial",	1,	quick_dial )		/* (SCO) Dial phone number #1, without progress detection */
#endif
#if ALL_ENTRIES || defined(exit_am_mode)
EXTRY('R','A', "rmam",	0,	exit_am_mode )		/* (SCO) Turn off automatic margins */
#endif
#if ALL_ENTRIES || defined(remove_clock)
EXTRY('R','C', "rmclk",	0,	remove_clock )		/* (SCO) Remove time-of-day clock */
#endif
#if ALL_ENTRIES || defined(req_for_input)
EXTRY('R','F', "rfi",	0,	req_for_input )		/* Send next input character (for ptys) */
#endif
ENTRY('R','I', "cuf",	PAD|1,	parm_right_cursor )	/* Move cursor right #1 spaces (PG) */
#if ALL_ENTRIES || defined(exit_xon_mode)
EXTRY('R','X', "rmxon",	0,	exit_xon_mode )		/* Turn off xon/xoff handshaking */
#endif
#if ALL_ENTRIES || defined(display_pc_char)
EXTRY('S','1', "dispc",	0,	display_pc_char )	/* (SCO) Displays PC character */
#endif /* CONFLICTS WITH: */
#if ALL_ENTRIES || defined(key_sf1)
EXTRY('S','1', "ksf1",	KEY,	key_sf1 )		/* (AIX) Special function key 1*/
#endif
#if ALL_ENTRIES || defined(key_sf2)
EXTRY('S','2', "ksf2",	KEY,	key_sf2 )		/* (AIX) Special function key 2*/
#endif
#if ALL_ENTRIES || defined(key_sf3)
EXTRY('S','3', "ksf3",	KEY,	key_sf3 )		/* (AIX) Special function key 3*/
#endif
#if ALL_ENTRIES || defined(enter_pc_charset_mode)
EXTRY('S','4', "smsc",	0,	enter_pc_charset_mode )	/* (SCO) Enables PC-scancode mode */
#endif /* CONFLICTS WITH: */
#if ALL_ENTRIES || defined(key_sf4)
EXTRY('S','4', "ksf4",	KEY,	key_sf4 )		/* (AIX) Special function key 4*/
#endif
#if ALL_ENTRIES || defined(exit_pc_charset_mode)
EXTRY('S','5', "rmsc",	0,	exit_pc_charset_mode )	/* (SCO) Disables PC-scancode mode */
#endif /* CONFLICTS WITH: */
#if ALL_ENTRIES || defined(key_sf5)
EXTRY('S','5', "ksf5",	KEY,	key_sf5 )		/* (AIX) Special function key 5*/
#endif
#if ALL_ENTRIES || defined(key_sf6)
EXTRY('S','6', "ksf6",	KEY,	key_sf6 )		/* (AIX) Special function key 6*/
#endif
#if ALL_ENTRIES || defined(key_sf7)
EXTRY('S','7', "ksf7",	KEY,	key_sf7 )		/* (AIX) Special function key 7*/
#endif
#if ALL_ENTRIES || defined(key_sf8)
EXTRY('S','8', "ksf8",	KEY,	key_sf8 )		/* (AIX) Special function key 8*/
#endif
#if ALL_ENTRIES || defined(key_sf9)
EXTRY('S','9', "ksf9",	KEY,	key_sf9 )		/* (AIX) Special function key 9*/
#endif
#if ALL_ENTRIES || defined(key_sf10)
EXTRY('S','0', "ksf10",	KEY,	key_sf10 )		/* (AIX) Special function key 10*/
#endif
#if ALL_ENTRIES || defined(enter_am_mode)
EXTRY('S','A', "smam",	0,	enter_am_mode )		/* Turn on automatic margins */
#endif
#if ALL_ENTRIES || defined(set_clock)
EXTRY('S','C', "sclk",	0/*?*/,	set_clock )		/* (SCO) Set time-of-day clock */
#endif
ENTRY('S','F', "indn",	PAD|1,	parm_index )		/* Scroll forward #1 lines (PG) */
ENTRY('S','R', "rin",	PAD|1,	parm_rindex )		/* Scroll backward #1 lines (PG) */
#if ALL_ENTRIES || defined(enter_xon_mode)
EXTRY('S','X', "smxon",	0,	enter_xon_mode )	/* Turn on xon/xoff handshaking */
#endif
#if ALL_ENTRIES || defined(set_background)
EXTRY('S','b', "setb",	1,	set_background )	/* (SCO) Set current background color #1 */
#endif
#if ALL_ENTRIES || defined(set_foreground)
EXTRY('S','f', "setf",	1,	set_foreground )	/* (SCO) Set current foreground color #1 */
#endif
#if ALL_ENTRIES || defined(tone)
EXTRY('T','O', "tone",	0,	tone )			/* (SCO) Select touch tone dialing */
#endif
ENTRY('U','P', "cuu",	PADM|1,	parm_up_cursor )	/* Move cursor up #1 lines (PG*) */
#if ALL_ENTRIES || defined(wait_tone)
EXTRY('W','A', "wait",	0,	wait_tone )		/* (SCO) Wait for dial tone */
#endif
#if ALL_ENTRIES || defined(goto_window)
EXTRY('W','G', "wingo",	1,	goto_window )		/* (SCO) Goto window #1.*/
#endif
#if ALL_ENTRIES || defined(xoff_character)
EXTRY('X','F', "xoffc",	0,	xoff_character )	/* X-off character */
#endif
#if ALL_ENTRIES || defined(xon_character)
EXTRY('X','N', "xonc",	0,	xon_character )		/* X-on character */
#endif
#if ALL_ENTRIES || defined(acs_chars)
EXTRY('a','c', "ascs",	0,	acs_chars )		/* Graphic charset pairs aAbBcC - def=vt100 */
#endif
ENTRY('a','e', "rmacs",	PAD,	exit_alt_charset_mode )	/* End alternate character set (P) */
ENTRY('a','l', "il1",	PADM,	insert_line )		/* Add new blank line (P*) */
ENTRY('a','s', "smacs",	PAD,	enter_alt_charset_mode ) /* Start alternate character set (P) */
ENTRY('b','l', "bel",	PAD,	bell )			/* Audible signal (bell) (P) */
ENTRY('b','t', "cbt",	PAD,	back_tab )		/* Back tab (P) */
#if ALL_ENTRIES || defined(box_chars_1)
EXTRY('b','x', "box1",	0,	box_chars_1 )		/* (AIX) Box characters primary set */
#endif
#if ALL_ENTRIES || defined(box_chars_2)
EXTRY('b','y', "box2",	0,	box_chars_2 )		/* (AIX) Box characters alternate set */
#endif
#if ALL_ENTRIES || defined(color_fg_0)
EXTRY('c','0', "colf0",	0,	color_fg_0 )		/* (AIX) Foreground color 0 white */
#endif
#if ALL_ENTRIES || defined(color_fg_1)
EXTRY('c','1', "colf1",	0,	color_fg_1 )		/* (AIX) Foreground color 1 red */
#endif
#if ALL_ENTRIES || defined(color_fg_2)
EXTRY('c','2', "colf2",	0,	color_fg_2 )		/* (AIX) Foreground color 2 green */
#endif
#if ALL_ENTRIES || defined(color_fg_3)
EXTRY('c','3', "colf3",	0,	color_fg_3 )		/* (AIX) Foreground color 3 brown */
#endif
#if ALL_ENTRIES || defined(color_fg_4)
EXTRY('c','4', "colf4",	0,	color_fg_4 )		/* (AIX) Foreground color 4 blue */
#endif
#if ALL_ENTRIES || defined(color_fg_5)
EXTRY('c','5', "colf5",	0,	color_fg_5 )		/* (AIX) Foreground color 5 magenta */
#endif
#if ALL_ENTRIES || defined(color_fg_6)
EXTRY('c','6', "colf6",	0,	color_fg_6 )		/* (AIX) Foreground color 6 cyan */
#endif
#if ALL_ENTRIES || defined(color_fg_7)
EXTRY('c','7', "colf7",	0,	color_fg_7 )		/* (AIX) Foreground color 7 black */
#endif
#if ALL_ENTRIES || defined(clr_bol)
ENTRY('c','b', "el1",	PAD,	clr_bol )		/* Clear to beginning of line, inclusive (P) */
#endif
ENTRY('c','d', "ed",	PADM,	clr_eos )		/* Clear to end of display (P*) */
ENTRY('c','e', "el",	PAD,	clr_eol )		/* Clear to end of line (P) */
ENTRY('c','h', "hpa",	PAD|1,	column_address )	/* Set cursor column (PG) */
ENTRY('c','l', "clear",	PADM,	clear_screen )		/* Clear screen and home cursor (P*) */
ENTRY('c','m', "cup",	PAD|2,	cursor_address )	/* Screen rel. cursor motion row #1 col #2 (PG) */
ENTRY('c','r', "cr",	PAD,	carriage_return )	/* Carriage return (P) */
ENTRY('c','s', "csr",	PAD|2,	change_scroll_region )	/* Change to lines #1 through #2 (vt100) (PG) */
ENTRY('c','t', "tbc",	PAD,	clear_all_tabs )	/* Clear all tab stops (P) */
ENTRY('c','v', "vpa",	PAD|1,	row_address )		/* Vertical position absolute (set row) (PG) */
#if ALL_ENTRIES || defined(color_bg_0)
EXTRY('d','0', "colb0",	0,	color_bg_0 )		/* (AIX) Background color 0 black */
#endif
#if ALL_ENTRIES || defined(color_bg_1)
EXTRY('d','1', "colb1",	0,	color_bg_1 )		/* (AIX) Background color 1 red */
#endif
#if ALL_ENTRIES || defined(color_bg_2)
EXTRY('d','2', "colb2",	0,	color_bg_2 )		/* (AIX) Background color 2 green */
#endif
#if ALL_ENTRIES || defined(color_bg_3)
EXTRY('d','3', "colb3",	0,	color_bg_3 )		/* (AIX) Background color 3 brown */
#endif
#if ALL_ENTRIES || defined(color_bg_4)
EXTRY('d','4', "colb4",	0,	color_bg_4 )		/* (AIX) Background color 4 blue */
#endif
#if ALL_ENTRIES || defined(color_bg_5)
EXTRY('d','5', "colb5",	0,	color_bg_5 )		/* (AIX) Background color 5 magenta */
#endif
#if ALL_ENTRIES || defined(color_bg_6)
EXTRY('d','6', "colb6",	0,	color_bg_6 )		/* (AIX) Background color 6 cyan */
#endif
#if ALL_ENTRIES || defined(color_bg_7)
EXTRY('d','7', "colb7",	0,	color_bg_7 )		/* (AIX) Background color 7 white */
#endif
ENTRY('d','c', "dch1",	PADM,	delete_character )	/* Delete character (P*) */
ENTRY('d','l', "dl1",	PADM,	delete_line )		/* Delete line (P*) */
ENTRY('d','m', "smdc",	0,	enter_delete_mode )	/* Delete mode (enter) */
ENTRY('d','o', "cud1",	0,	cursor_down )		/* Down one line */
ENTRY('d','s', "dsl",	0,	dis_status_line )	/* Disable status line */
#if ALL_ENTRIES || defined(ena_acs)
EXTRY('e','A', "enacs",	0,	ena_acs )		/* Enable alternate character set */
#endif
ENTRY('e','c', "ech",	PAD|1,	erase_chars )		/* Erase #1 characters (PG) */
ENTRY('e','d', "rmdc",	0,	exit_delete_mode )	/* End delete mode */
ENTRY('e','i', "rmir",	0,	exit_insert_mode )	/* End insert mode */
#if ALL_ENTRIES || defined(font_0)
EXTRY('f','0', "font0",	0,	font_0 )		/* (AIX) Select font 0 */
#endif
#if ALL_ENTRIES || defined(font_1)
EXTRY('f','1', "font1",	0,	font_1 )		/* (AIX) Select font 1 */
#endif
#if ALL_ENTRIES || defined(font_2)
EXTRY('f','2', "font2",	0,	font_2 )		/* (AIX) Select font 2 */
#endif
#if ALL_ENTRIES || defined(font_3)
EXTRY('f','3', "font3",	0,	font_3 )		/* (AIX) Select font 3 */
#endif
#if ALL_ENTRIES || defined(font_4)
EXTRY('f','4', "font4",	0,	font_4 )		/* (AIX) Select font 4 */
#endif
#if ALL_ENTRIES || defined(font_5)
EXTRY('f','5', "font5",	0,	font_5 )		/* (AIX) Select font 5 */
#endif
#if ALL_ENTRIES || defined(font_6)
EXTRY('f','6', "font6",	0,	font_6 )		/* (AIX) Select font 6 */
#endif
#if ALL_ENTRIES || defined(font_7)
EXTRY('f','7', "font7",	0,	font_7 )		/* (AIX) Select font 7 */
#endif
ENTRY('f','f', "ff",	PADM,	form_feed )		/* Hardcopy terminal page eject (P*) */
#if ALL_ENTRIES || defined(flash_hook)
EXTRY('f','h', "hook",	0,	flash_hook )		/* (SCO) Flash the switch hook */
#endif
ENTRY('f','s', "fsl",	0,	from_status_line )	/* Return from status line */
ENTRY('h','d', "hd",	0,	down_half_line )	/* Half-line down (forward 1/2 linefeed) */
ENTRY('h','o', "home",	PAD,	cursor_home )		/* Home cursor (if no cup) (P) */
ENTRY('h','u', "hu",	0,	up_half_line )		/* Half-line up (reverse 1/2 linefeed) */
ENTRY('i','1', "is1",	0,	init_1string )		/* Terminal initialization string */
ENTRY('i','2', "is3",	0,	init_3string )		/* Terminal initialization string */
ENTRY('i','P', "iprog",	0,	init_prog )		/* Path name of program for initialization */
ENTRY('i','c', "ich1",	PADM,	insert_character )	/* Insert character (P*) */
ENTRY('i','f', "if",	0,	init_file )		/* Name of initialization file */
ENTRY('i','m', "smir",	0,	enter_insert_mode )	/* Insert mode (enter); */
ENTRY('i','p', "ip",	PADM,	insert_padding )	/* Insert pad after character inserted (P*) */
ENTRY('i','s', "is2",	0,	init_2string )		/* Terminal initialization string */
ENTRY('k','0', "kf0",	KEY,	key_f0 )		/* Sent by function key f0 */
ENTRY('k','1', "kf1",	KEY,	key_f1 )		/* Sent by function key f1 */
ENTRY('k','2', "kf2",	KEY,	key_f2 )		/* Sent by function key f2 */
ENTRY('k','3', "kf3",	KEY,	key_f3 )		/* Sent by function key f3 */
ENTRY('k','4', "kf4",	KEY,	key_f4 )		/* Sent by function key f4 */
ENTRY('k','5', "kf5",	KEY,	key_f5 )		/* Sent by function key f5 */
ENTRY('k','6', "kf6",	KEY,	key_f6 )		/* Sent by function key f6 */
ENTRY('k','7', "kf7",	KEY,	key_f7 )		/* Sent by function key f7 */
ENTRY('k','8', "kf8",	KEY,	key_f8 )		/* Sent by function key f8 */
ENTRY('k','9', "kf9",	KEY,	key_f9 )		/* Sent by function key f9 */
ENTRY('k',';', "kf10",	KEY,	key_f10 )		/* Sent by function key f10 */
ENTRY('k','A', "kil1",	KEY,	key_il )		/* Sent by insert line */
#if ALL_ENTRIES || defined(key_back_tab)
EXTRY('k','B', "kbtab",	KEY,	key_back_tab )		/* Backtab key */
#endif
ENTRY('k','C', "kclr",	KEY,	key_clear )		/* Sent by clear screen or erase key */
ENTRY('k','D', "kdch1",	KEY,	key_dc )		/* Sent by delete character key */
ENTRY('k','E', "kel",	KEY,	key_eol )		/* Sent by clear-to-end-of-line key */
ENTRY('k','F', "kind",	KEY,	key_sf )		/* Sent by scroll-forward/down key */
ENTRY('k','H', "kll",	KEY,	key_ll )		/* Sent by home-down key */
ENTRY('k','I', "kich1",	KEY,	key_ic )		/* Sent by ins char/enter ins mode key */
#if ALL_ENTRIES || defined(key_action)
EXTRY('k','J', "kact",	KEY,	key_action )		/* (AIX) Sent by the action key */
#endif
ENTRY('k','L', "kdl1",	KEY,	key_dl )		/* Sent by delete line key */
ENTRY('k','M', "krmir",	KEY,	key_eic )		/* Sent by rmir or smir in insert mode */
ENTRY('k','N', "knp",	KEY,	key_npage )		/* Sent by next-page key */
ENTRY('k','P', "kpp",	KEY,	key_ppage )		/* Sent by previous-page key */
#if ALL_ENTRIES || defined(key_quit)
EXTRY('k','Q', "kquit",	KEY,	key_quit )		/* (AIX) Quit key */
#endif
ENTRY('k','R', "kri",	KEY,	key_sr )		/* Sent by scroll-backward/up key */
ENTRY('k','S', "ked",	KEY,	key_eos )		/* Sent by clear-to-end-of-screen key */
ENTRY('k','T', "khts",	KEY,	key_stab )		/* Sent by set-tab key */
#if ALL_ENTRIES || defined(key_select)
EXTRY('k','U', "ksel",	KEY,	key_select )		/* (AIX) Select key */
#endif
#if ALL_ENTRIES || defined(key_prev_pane)
EXTRY('k','V', "kppn",	KEY,	key_prev_pane )		/* (AIX) Previous pane key */
#endif
#if ALL_ENTRIES || defined(key_command_pane)
EXTRY('k','W', "kcpn",	KEY,	key_command_pane )	/* (AIX) Command pane key */
#endif
#if ALL_ENTRIES || defined(key_scroll_right)
EXTRY('k','Z', "kscr",	KEY,	key_scroll_right )	/* Scroll right */
#endif
ENTRY('k','a', "ktbc",	KEY,	key_catab )		/* Sent by clear-all-tabs key */
ENTRY('k','b', "kbs",	KEY,	key_backspace )		/* Sent by backspace key */
#if ALL_ENTRIES || defined(key_command)
EXTRY('k','c', "kcmd",	KEY,	key_command )		/* (AIX) Command request key */
#endif
ENTRY('k','d', "kcud1",	KEY,	key_down )		/* Sent by terminal down arrow key */
ENTRY('k','e', "rmkx",	0,	keypad_local )		/* Out of "keypad transmit" mode */
ENTRY('k','h', "khome",	KEY,	key_home )		/* Sent by home key */
#if ALL_ENTRIES || defined(key_do)
EXTRY('k','i', "kdo",	KEY,	key_do )		/* (AIX) Do request key */
#endif
ENTRY('k','l', "kcub1",	KEY,	key_left )		/* Sent by terminal left arrow key */
#if ALL_ENTRIES || defined(key_newline)
EXTRY('k','n', "knl",	KEY,	key_newline )		/* (AIX) Newline key */
#endif
#if ALL_ENTRIES || defined(key_tab)
EXTRY('k','o', "ktab",	KEY,	key_tab )		/* (AIX) Tab key */
#endif
#if ALL_ENTRIES || defined(key_prev_cmd)
EXTRY('k','p', "kpcmd",	KEY,	key_prev_cmd )		/* (AIX) Previous command key */
#endif
#if ALL_ENTRIES || defined(key_help)
EXTRY('k','q', "khlp",	KEY,	key_help )		/* (AIX) Help key */
#endif
ENTRY('k','r', "kcuf1",	KEY,	key_right )		/* Sent by terminal right arrow key */
ENTRY('k','s', "smkx",	0,	keypad_xmit )		/* Put terminal in "keypad transmit" mode */
ENTRY('k','t', "kctab",	KEY,	key_ctab )		/* Sent by clear-tab key */
ENTRY('k','u', "kcuu1",	KEY,	key_up )		/* Sent by terminal up arrow key */
#if ALL_ENTRIES || defined(key_next_pane)
EXTRY('k','v', "knpn",	KEY,	key_next_pane )		/* (AIX) Next pane key */
#endif
#if ALL_ENTRIES || defined(key_end)
EXTRY('k','w', "kend",	KEY,	key_end )		/* (AIX) End key */
#endif
#if ALL_ENTRIES || defined(key_scroll_left)
EXTRY('k','z', "kscl",	KEY,	key_scroll_left )	/* (AIX) Scroll left */
#endif
ENTRY('l','0', "lf0",	0,	lab_f0 )		/* Labels on function key f0 if not f0 */
ENTRY('l','1', "lf1",	0,	lab_f1 )		/* Labels on function key f1 if not f1 */
ENTRY('l','2', "lf2",	0,	lab_f2 )		/* Labels on function key f2 if not f2 */
ENTRY('l','3', "lf3",	0,	lab_f3 )		/* Labels on function key f3 if not f3 */
ENTRY('l','4', "lf4",	0,	lab_f4 )		/* Labels on function key f4 if not f4 */
ENTRY('l','5', "lf5",	0,	lab_f5 )		/* Labels on function key f5 if not f5 */
ENTRY('l','6', "lf6",	0,	lab_f6 )		/* Labels on function key f6 if not f6 */
ENTRY('l','7', "lf7",	0,	lab_f7 )		/* Labels on function key f7 if not f7 */
ENTRY('l','8', "lf8",	0,	lab_f8 )		/* Labels on function key f8 if not f8 */
ENTRY('l','9', "lf9",	0,	lab_f9 )		/* Labels on function key f9 if not f9 */
ENTRY('l','a', "lf10",	0,	lab_f10 )		/* Labels on function key f10 if not f10 */
ENTRY('l','e', "cub1",	PAD,	cursor_left )		/* Move cursor left one space */
ENTRY('l','l', "ll",	0,	cursor_to_ll )		/* Last line, first column (if no cup) */
ENTRY('m','b', "blink",	0,	enter_blink_mode )	/* Turn on blinking */
ENTRY('m','d', "bold",	0,	enter_bold_mode )	/* Turn on bold (extra bright) mode */
ENTRY('m','e', "sgr0",	0,	exit_attribute_mode )	/* Turn off all attributes */
ENTRY('m','h', "dim",	0,	enter_dim_mode )	/* Turn on half-bright mode */
ENTRY('m','k', "invis",	0,	enter_secure_mode )	/* Turn on blank mode (chars invisible) */
#if ALL_ENTRIES || defined(memory_lock)
EXTRY('m','l', "meml",	0,	memory_lock )		/* (AIX) Lock memory above cursor (o) */
#endif
ENTRY('m','m', "smm",	0,	meta_on )		/* Turn on "meta mode" (8th bit) */
ENTRY('m','o', "rmm",	0,	meta_off )		/* Turn off "meta mode" */
ENTRY('m','p', "prot",	0,	enter_protected_mode )	/* Turn on protected mode */
ENTRY('m','r', "rev",	0,	enter_reverse_mode )	/* Turn on reverse video mode */
#if ALL_ENTRIES || defined(memory_unlock)
EXTRY('m','u', "memu",	0,	memory_unlock )		/* (AIX) Turn memory lock off (o) */
#endif
ENTRY('n','d', "cuf1",	0,	cursor_right )		/* Non-destructive space (cursor right) */
ENTRY('n','w', "nel",	PAD,	newline )		/* Newline (behaves like cr followed by lf) */
#if ALL_ENTRIES || defined(orig_colors)
EXTRY('o','c', "oc",	0,	orig_colors )		/* (SCO) Set all colors to the original ones */
#endif
#if ALL_ENTRIES || defined(orig_pair)
EXTRY('o','p', "op",	0,	orig_pair )		/* (SCO) Set default color-pair to the original */
#endif
ENTRY('p','O', "mc5p",	1,	prtr_non )		/* Turn on the printer for #1 bytes */
ENTRY('p','c', "pad",	0,	pad_char )		/* Pad character (rather than null) */
ENTRY('p','f', "mc4",	0,	prtr_off )		/* Turn off the printer */
ENTRY('p','k', "pfkey",	2,	pkey_key )		/* Prog funct key #1 to type string #2 */
ENTRY('p','l', "pfloc",	2,	pkey_local )		/* Prog funct key #1 to execute string #2 */
#if ALL_ENTRIES || defined(plab_norm)
ENTRY('p','n', "pln",	2,	plab_norm )		/* Prog label #1 to show string #2 */
#endif
ENTRY('p','o', "mc5",	0,	prtr_on )		/* Turn on the printer */
ENTRY('p','s', "mc0",	0,	print_screen )		/* Print contents of the screen */
ENTRY('p','x', "pfx",	2,	pkey_xmit )		/* Prog funct key #1 to xmit string #2 */
ENTRY('r','1', "rs1",	0,	reset_1string )		/* Reset terminal completely to sane modes */
ENTRY('r','2', "rs2",	0,	reset_2string )		/* Reset terminal completely to sane modes */
ENTRY('r','3', "rs3",	0,	reset_3string )		/* Reset terminal completely to sane modes */
#if ALL_ENTRIES || defined(char_padding)
EXTRY('r','P', "mp",	PAD,	char_padding )		/* like `ip' but in replace mode */
#endif
ENTRY('r','c', "rc",	PAD,	restore_cursor )	/* Restore cursor to position of last sc (P) */
ENTRY('r','f', "rf",	0,	reset_file )		/* Name of file containing reset string */
ENTRY('r','p', "rep",	PADM|2,	repeat_char )		/* Repeat char #1 #2 times. (PG*) */
ENTRY('s','a', "sgr",	PAD|9,	set_attributes )	/* Define the video attributes (PG9) */
ENTRY('s','c', "sc",	PAD,	save_cursor )		/* Save cursor position (P) */
ENTRY('s','e', "rmso",	0,	exit_standout_mode )	/* End stand out mode */
ENTRY('s','f', "ind",	PAD,	scroll_forward )	/* Scroll text up (P) */
ENTRY('s','o', "smso",	0,	enter_standout_mode )	/* Begin stand out mode */
#if ALL_ENTRIES || defined(set_color_pair)
EXTRY('s','p', "setp",	1,	set_color_pair )	/* (SCO) Set current color pair #1.*/
#endif
ENTRY('s','r', "ri",	PAD,	scroll_reverse )	/* Scroll text down (P) */
ENTRY('s','t', "hts",	0,	set_tab )		/* Set a tab in all rows, current column */
ENTRY('t','a', "ht",	PAD,	tab )			/* Tab to next 8 space hardware tab stop */
ENTRY('t','e', "rmcup",	0,	exit_ca_mode )		/* String to end programs that use cup */
ENTRY('t','i', "smcup",	0,	enter_ca_mode )		/* String to begin programs that use  cup */
ENTRY('t','s', "tsl",	1,	to_status_line )	/* Go to status line, column #1 */
#if ALL_ENTRIES || defined(user0)
EXTRY('u','0', "u0",	0,	user0 )			/* User string 0 */
#endif
#if ALL_ENTRIES || defined(user1)
EXTRY('u','1', "u1",	0,	user1 )			/* User string 1 */
#endif
#if ALL_ENTRIES || defined(user2)
EXTRY('u','2', "u2",	0,	user2 )			/* User string 2 */
#endif
#if ALL_ENTRIES || defined(user3)
EXTRY('u','3', "u3",	0,	user3 )			/* User string 3 */
#endif
#if ALL_ENTRIES || defined(user4)
EXTRY('u','4', "u4",	0,	user4 )			/* User string 4 */
#endif
#if ALL_ENTRIES || defined(user5)
EXTRY('u','5', "u5",	0,	user5 )			/* User string 5 */
#endif
#if ALL_ENTRIES || defined(user6)
EXTRY('u','6', "u6",	0,	user6 )			/* User string 6 */
#endif
#if ALL_ENTRIES || defined(user7)
EXTRY('u','7', "u7",	0,	user7 )			/* User string 7 */
#endif
#if ALL_ENTRIES || defined(user8)
EXTRY('u','8', "u8",	0,	user8 )			/* User string 8 */
#endif
#if ALL_ENTRIES || defined(user9)
EXTRY('u','9', "u9",	0,	user9 )			/* User string 9 */
#endif
ENTRY('u','c', "uc",	0,	underline_char )	/* Underscore one char and move past it */
ENTRY('u','e', "rmul",	0,	exit_underline_mode )	/* End underscore mode */
ENTRY('u','p', "cuu1",	0,	cursor_up )		/* Upline (cursor up) */
ENTRY('u','s', "smul",	0,	enter_underline_mode )	/* Start underscore mode */
ENTRY('v','b', "flash",	0,	flash_screen )		/* Visible bell (may not move cursor) */
ENTRY('v','e', "cnorm",	0,	cursor_normal )		/* Make cursor appear normal (undo vs/vi) */
ENTRY('v','i', "civis",	0,	cursor_invisible )	/* Make cursor invisible */
ENTRY('v','s', "cvvis",	0,	cursor_visible )	/* Make cursor very visible */
ENTRY('w','i', "wind",	4,	set_window )		/* Current window is lines #1-#2 cols #3-#4 */
#if ALL_ENTRIES || defined(appl_defined_str)
EXTRY('z','a', "apstr",	0,	appl_defined_str )	/* (AIX) Application defined terminal string*/
#endif

#endif /* STRINGS */

#undef TABLE
#undef FLAGS
#undef NUMBERS
#undef STRINGS

#endif /* TABLE */
